@@ -452,6 +452,7 @@ parse_format_list (void)
   /* Get the next format item */
  format_item:
   t = format_lex ();
+ format_item_1:
   switch (t)
     {
     case FMT_POSINT:
@@ -575,6 +576,8 @@ parse_format_list (void)
 
     case FMT_DOLLAR:
       get_fnode (&head, &tail, FMT_DOLLAR);
+      tail->repeat = 1;
+      notify_std (GFC_STD_GNU, "Extension: $ descriptor");
       goto between_desc;
 
     case FMT_T:
@@ -853,8 +856,8 @@ parse_format_list (void)
       goto finished;
 
     default:
-      error = "Missing comma in format";
-      goto finished;
+      /* Assume a missing comma, this is a GNU extension */
+      goto format_item_1;
     }
 
   /* Optional comma is a weird between state where we've just finished
